#include "../std.h"

extern "C" {
#include <globus_replica_catalog.h>
#include <globus_io.h>
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
#include <globus_rls_client.h>
#endif
}

#include "../config/config_map.h"
#include "../misc/url_options.h"
#include "../misc/checksum.h"
#include "../misc/globus_modules.h"
#include "../misc/log_time.h"

#include "datapoint.h"

DataPoint::~DataPoint(void) {
}

DataPoint::DataPoint(const char* url_) {
  is_valid=false; tries_left=5;
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
  if((!common_mod.active()) || (!io_mod.active()) || (!rls_mod.active())) return;
#endif
  if((!error_mod.active()) || (!gsi_gssapi_mod.active()) || (!replica_mod.active())) return;
  if((url_ == NULL) || (url_[0]==0)) return;
  common_url_options="";
  is_valid=true; url=url_;
  locations.clear(); location=locations.end();
  is_file=false; is_ftp=false; is_bbftp=false; is_http=false;
  is_rc=false; is_magda=false; is_rls=false;
#ifndef SKIP_HTTPG
  is_se=false;
#endif
  is_secure=false; is_resolved=true; 
  meta_size_valid=false; meta_checksum_valid=false;
  meta_created_valid=false; meta_validtill_valid=false;
  is_metaexisting=false;
/*  meta_valid=false; */
  if(strncasecmp("file:/",url_,6) == 0) {
    is_file=true; is_secure=false;
    locations.push_back(url_); location=locations.begin();
  }
  else if(strncasecmp("rc://",url_,5) == 0) {
    is_rc=true; is_secure=true; is_resolved=false;
    if(!process_meta_url_rc(url_,rc_url,rc_lfn,locations)) {
      is_valid=false;
    };
    location=locations.begin(); rc_mgr=NULL;
  }
  else if(strncasecmp("rls://",url_,6) == 0) {
    is_rls=true; is_secure=true; is_resolved=false;
    if(!process_meta_url_rls(url_,rls_url,rls_lfn,locations)) {
      is_valid=false;
    };
    location=locations.begin();
  }
#ifndef SKIP_HTTPG
  else if(strncasecmp("se://",url_,5) == 0) {
    is_se=true; is_secure=true; is_resolved=false;
    if(!process_meta_url_se(url_,se_url,se_lfn,locations)) {
cerr<<"DataPoint::DataPoint: process_meta_url_se failed"<<endl;
      is_valid=false;
    };
cerr<<"DataPoint::DataPoint: se_lfn: "<<se_lfn<<endl;
    location=locations.begin();
  }
#endif
  else if(strncasecmp("bbftp://",url_,8) == 0) {
    is_bbftp=true; is_secure=false;
    locations.push_back(url_); location=locations.begin();
  }
  else if(strncasecmp("gsibbftp://",url_,11) == 0) {
    is_bbftp=true; is_secure=true;
    locations.push_back(url_); location=locations.begin();
  }
  else if(strncasecmp("magda://",url_,8) == 0) {
    is_magda=true; is_secure=false;
    locations.push_back(url_); location=locations.begin();
  }
  else if(strncasecmp("ftp://",url_,6) == 0) {
    is_ftp=true; is_secure=false;
    locations.push_back(url_); location=locations.begin();
  }
  else if(strncasecmp("gsiftp://",url_,9) == 0) {
    is_ftp=true; is_secure=true;
    locations.push_back(url_); location=locations.begin();
  }
  else if(strncasecmp("http://",url_,7) == 0) {
    is_http=true; is_secure=false;
    locations.push_back(url_); location=locations.begin();
  }
  else if(strncasecmp("https://",url_,8) == 0) {
    is_http=true; is_secure=true;
    locations.push_back(url_); location=locations.begin();
  }
  else {
    odlog(0)<<"DataPoint: unrecognized protocol"<<endl;
    is_valid=false; location=locations.begin();
  };
}

bool DataPoint::get_info(DataPoint::FileInfo &fi) {
  if(!meta()) { return true; }; // not a meta url
  if(!meta_resolve(true)) { return false; }; // resolution failure
  fi.name=lfn();
  for(list<DataPoint::Location>::iterator i = locations.begin();
                                              i!=locations.end();++i) {
    fi.urls.push_back(i->url);
  };  
  if(meta_size_valid) { fi.size=meta_size_; fi.size_available=true; };
  if(meta_checksum_valid) { fi.checksum=meta_checksum_; fi.checksum_available=true; };
  if(meta_created_valid) { fi.created=meta_created_; fi.created_available=true; };
  if(meta_validtill_valid) { fi.valid=meta_validtill_; fi.valid_available=true; };
  fi.type=DataPoint::FileInfo::file_type_file;
  return false;
}

bool DataPoint::list_files(list<DataPoint::FileInfo> &files,bool resolve) {
  if(!meta()) return true;
  if(strncasecmp("rc://",url.c_str(),5) == 0) {
    odlog(2)<<"list: RC url"<<endl;
    return list_files_rc(files,resolve);
  }
  else if(strncasecmp("rls://",url.c_str(),6) == 0) {
    odlog(2)<<"list: RLS url"<<endl;
    return list_files_rls(files,resolve);
#ifndef SKIP_HTTPG
  } else if(strncasecmp("se://",url.c_str(),5) == 0) {
    odlog(2)<<"list: SE url"<<endl;
    return list_files_se(files,resolve);
#endif
  };
  return false;
}

/* perform resolve operation, which can take long time */
bool DataPoint::meta_resolve(bool source,const UrlMap &maps) {
  if(is_resolved) return true;
  bool res=meta_resolve(source);
  if(!res) return false;
  sort(maps); location=locations.begin();
  return true;
}

bool DataPoint::meta_resolve(bool source) {
  if(is_resolved) return true;
  if(!meta()) return true;
  odlog(2)<<"meta_resolve: url: "<<url<<endl;
  if(strncasecmp("rc://",url.c_str(),5) == 0) {
    odlog(2)<<"meta_resolve: RC url"<<endl;
    return meta_resolve_rc(source);
  }
  else if(strncasecmp("rls://",url.c_str(),6) == 0) {
    odlog(2)<<"meta_resolve: RLS url"<<endl;
    return meta_resolve_rls(source);
#ifndef SKIP_HTTPG
  }
  else if(strncasecmp("se://",url.c_str(),5) == 0) {
    odlog(2)<<"meta_resolve: SE url"<<endl;
    return meta_resolve_se(source);
#endif
  };


  return false;
}

bool DataPoint::meta_preregister(bool replication,bool force) {
  if(!meta()) return true;
  odlog(2)<<"meta_preregister"<<endl;
  if(strncasecmp("rc://",url.c_str(),5) == 0) {
    return meta_preregister_rc(replication,force);
  }
  else if(strncasecmp("rls://",url.c_str(),6) == 0) {
    return meta_preregister_rls(replication,force);
  };
  return true;
}

bool DataPoint::meta_postregister(bool replication,bool failure) {
  if(!meta()) return true;
  odlog(2)<<"meta_postregister"<<endl;
  if(strncasecmp("rc://",url.c_str(),5) == 0) {
    odlog(2)<<"meta_postregister: is rc"<<endl;
    return meta_postregister_rc(replication,failure);
  }
  else if(strncasecmp("rls://",url.c_str(),6) == 0) {
    odlog(2)<<"meta_postregister: is rls"<<endl;
    return meta_postregister_rls(replication,failure);
  };
  return true;
}

bool DataPoint::meta_unregister(bool all) {
  if(!meta()) return true;
  if(strncasecmp("rc://",url.c_str(),5) == 0) {
/* not fully implemented yet - all=true will remove logical file only */
    odlog(2)<<"meta_unregister: is rc"<<endl;
    return meta_unregister_rc(all);
  }
  else if(strncasecmp("rls://",url.c_str(),6) == 0) {
    odlog(2)<<"meta_unregister: is rls"<<endl;
    return meta_unregister_rls(all);
  };
  return true;
}

bool DataPoint::meta_preunregister(bool replication) {
  if(!meta()) return true;
  if(strncasecmp("rc://",url.c_str(),5) == 0) {
    odlog(2)<<"meta_preunregister: is rc"<<endl;
    return meta_preunregister_rc(replication);
  }
  else if(strncasecmp("rls://",url.c_str(),6) == 0) {
    odlog(2)<<"meta_preunregister: is rls"<<endl;
    return meta_preunregister_rls(replication);
  };
  return true;
}

bool DataPoint::have_locations(void) {
  return (locations.size() != 0);
}

bool DataPoint::have_location(void) {
  if(!is_resolved) {
    if(tries_left <= 0) return false;
    return true;
  };
  if(location == locations.end()) return false;
  return true;
}

bool DataPoint::next_location(void) {
  if(!is_resolved) {
    tries_left--;
    if(tries_left <= 0) return false;
    return true;
  };
  if(location == locations.end()) return false;
  ++location;
  if(location == locations.end()) {
    tries_left--;
    if(tries_left <= 0) return false;
    location=locations.begin();
  };
  return true;
}

bool DataPoint::remove_location(void) {
  if(location == locations.end()) return false;
  location=locations.erase(location);
  return true;
}

string DataPoint::canonic_url(void) const {
  string tmp = url;
  if(::canonic_url(tmp) != 0) tmp="";
  return tmp;
}

string DataPoint::base_url(void) const {
  return url;
}

ostream& operator<<(ostream& o,const DataPoint &point) {
  if(!point.is_valid) { o<<"<invalid>"; return o; };
  if(point.is_rc || point.is_magda) {
    if(point.is_resolved) {
      if(point.locations.end() != point.location) {
        o<<point.url<<"["<<point.location->url<<"]";
        return o;
      };
    };   
  };  
  o<<point.url;
  return o;
}

/* put local urls first and shuffle rest */
bool DataPoint::sort(const UrlMap &maps) {
  list<Location>::iterator ii = locations.begin();
  int nn = 0;
  for(list<Location>::iterator i=locations.begin();i!=locations.end();) {
    string c_url = i->url;
    ::canonic_url(c_url);
    if(maps.local(c_url)) {  /* move to first place */
      if(i == ii) { ++ii; ++i; nn++; continue; };
      locations.insert(ii,*i);
      if(i == location) location=locations.begin();
      i=locations.erase(i);
      nn++;
    }
    else {
      ++i;
    };
  };
  nn=locations.size()-nn;
  if(nn<=1) return true;
  srandom(time(NULL));
  for(;nn>1;) {
    int r = random()/2;
    int n=r/((RAND_MAX/2+1)/nn);
    list<Location>::iterator i=ii;
    for(;n;n--) { ++i; };
    if(i == locations.end()) { nn--; ++ii; continue; };
    if(i == ii) { nn--; ++ii; continue; };
    locations.insert(ii,*i);
    if(i == location) location=locations.begin();
    i=locations.erase(i);
    nn--;
  };
  return true;
}

bool DataPoint::map(const UrlMap &maps) {
  for(list<Location>::iterator i=locations.begin();i!=locations.end();) {
    if(maps.map(i->url)) {  /* move to first place */
      locations.push_front(*i);
      if(i == location) location=locations.begin();
      i=locations.erase(i);
    }
    else {
      ++i;
    };
  };
  return true;
}

bool DataPoint::add_location(const char* meta_loc,const char* loc) {
  if(!meta()) return false;
  odlog(2)<<"add_location: meta location: "<<meta_loc<<endl;
  odlog(2)<<"add_location: location: "<<loc<<endl;
  for(list<Location>::iterator i = locations.begin();i!=locations.end();++i) {
    if(i->meta == meta_loc) return true;
  };
  locations.insert(locations.end(),Location(meta_loc,loc,false));
  return true;
}

