/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/

#include <sys/types.h>
#include <dirent.h>

#include <voms_apic.h>

extern int retrieve(X509 *, STACK_OF(X509) *, int, char **, int *, char **, int *, char **, int *, int *);

void 
VOMS_Destroy(struct vomsdata *vd)
{
  int i = 0, j=0;
  struct voms *v;
  struct data *d;

  if (vd) {
    if (vd->data) {
      while (v = vd->data[i++]) {
	if (v->std && (v->type == TYPE_STD)) {
	  while (d = v->std[j++])
	    free(d);
	  free (v->std);
	}
	free(v->reserved);
	free(v);
      }
      free(vd->data);
    }
    if (vd->workvo)
      free (vd->workvo);
    if (vd->extra_data)
      free (vd->extra_data);
    free(vd);
  }
}

  
struct vomsdata *
VOMS_Init(char *v, char *c)
{
  int result = 1;
  struct vomsdata *tmp;
  DIR *vdir, *cdir;

  if (!v)
    if (!(v = getenv("X509_VOMS_DIR")))
      v = "/etc/grid-security/vomsdir";

  if (!c)
    if (!(c = getenv("X509_CERT_DIR")))
      c = "/etc/grid-security/certificates";

  vdir = opendir(v);
  cdir = opendir(c);

  if (!vdir || !cdir)
    result = 0;
    
  if (cdir)
    (void)closedir(cdir);
  if (vdir)
    (void)closedir(vdir);

  if (!result)
    return NULL;

  if (tmp = (struct vomsdata *)malloc(sizeof(struct vomsdata))) {
    tmp->cdir = c;
    tmp->vdir = v;
    tmp->extra_data = tmp->workvo = NULL;
    tmp->data = NULL;
  }
  return tmp;

}

int 
VOMS_Contact(char *hostname, int port, char *servsubject, char *command, struct vomsdata *vd, int *error)
{
  char *buffer;
  int buflen;

  *error = VERR_NOSOCKET;

  if (net_connect(hostname, port, servsubject, command, strlen(command), &buffer, &buflen)) {
    vd->extra_data = vd->workvo = NULL;
    vd->extralen = vd->volen = 0;
    return VOMS_Import(buffer, buflen, vd, error);
  }

  return 0;
}

int 
VOMS_Retrieve(X509 *cert, STACK_OF(X509) *chain, int how, struct vomsdata *vd, int *error)
{
  char *buffer;
  char *vo;
  char *file;
  int buflen, volen, filelen;

  if (retrieve(cert, chain, how, &buffer, &buflen, &vo, &volen, &file, &filelen, error)) {
    vd->workvo = vo;
    vd->extra_data = file;
    vd->volen = volen;
    vd->extralen = filelen;
    return VOMS_Import(buffer, buflen, vd, error);
  }
  return 0;
}

extern int verify(char *message, int len, struct vomsdata *voms, int *error);

int VOMS_Import(char *buffer, int buflen, struct vomsdata *vd, int *error)
{
  return verify(buffer, buflen, vd, error);
}

int getMajorVersionNumber(void) {return 1;}
int getMinorVersionNumber(void) {return 0;}
int getPatchVersionNumber(void) {return 0;}
