/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/
#include <iostream>
#include <voms_api.h>

#include <openssl/pem.h>
#include <openssl/x509.h>

static string retmsg[] = { "VERR_NONE", "VERR_NOSOCKET", "VERR_NOIDENT", "VERR_COMM", 
			"VERR_PARAM", "VERR_NOEXT", "VERR_NOINIT",
			"VERR_TIME", "VERR_IDCHECK", "VERR_EXTRAINFO",
			"VERR_FORMAT", "VERR_NODATA", "VERR_PARSE",
			"VERR_DIR", "VERR_SIGN", "VERR_SERVER", 
			"VERR_MEM", "VERR_VERIFY", "VERR_IDENT",
			"VERR_TYPE" };

void print(vomsdata &d)
{
  vector<voms> v = d.data;
  int k = 0;

  for (voms *i=v.begin(); i != v.end(); i++) {
    cout << ++k << " ********************************************" << endl;
    cout << "SIGLEN: " << i->siglen << endl << "USER:" << i->user << endl
	 << "UCA: " << i->userca << endl << "SERVER: " << i->server << endl
	 << "SCA: " << i->serverca << endl << "VO: " << i->voname << endl
	 << "URI: " << i->uri << endl << "DATE1: " << i->date1 << endl
	 << "DATE2: " << i->date2 << endl;

    switch (i->type) {
    case TYPE_NODATA:
      cout << "NO DATA" << endl;
      break;
    case TYPE_CUSTOM:
      cout << i->custom << endl;
      break;
    case TYPE_STD:
      for (data *j = i->std.begin(); j != i->std.end(); j++)
	cout << "GROUP: " << j->group << endl
	     << "ROLE: " << j->role << endl
	     << "CAP: " << j->cap << endl;
      break;
    }
  }
  cout << "WORKVO: " << d.workvo << endl
       << "EXTRA: " << d.extra_data << endl << flush;
}

int main(int argc, char *argv[])
{
  vomsdata d("/home/marotta/API/Auth/voms/api/cert");
  verror_type err;
  bool res;

  /*
   * Initializes directory info.
   */

  cerr << "TEST 1" << endl;
  /*
   * Contact two different vomses and collect information from both.
   */
  res = d.Contact("aaa-test.cnaf.infn.it",15000,"/C=IT/O=INFN/OU=cas server/L=Bologna/CN=cas/aaa-test.cnaf.infn.it/Email=Vincenzo.Ciaschini@cnaf.infn.it", "S1");
  res = d.Contact("datatag6.cnaf.infn.it",15000,"/C=IT/O=INFN/OU=cas server/L=Bologna/CN=cas/aaa-test.cnaf.infn.it/Email=Vincenzo.Ciaschini@cnaf.infn.it", "A");

  if (res)
    print(d);
  else
    cerr << "ERROR!" << endl << flush;
  if (!res) {
    cerr << "err: " << retmsg[d.error] << endl;
  }

  cerr << "TEST 1.5" << endl;

  /*
   * Unparse info into its original format.
   */
  string binary;
  res = d.Export(binary);

  if (res)
    cout << "------------------" << endl << binary << "-------" << endl << flush;
  else
    cerr << "Error!" << endl << flush;
  if (!res) {
    cerr << "err: " << retmsg[d.error] << endl;
  }

  /*
   * Now convert it back into readable format.
   */
  vomsdata v("/home/marotta/API/Auth/voms/api/cert","/etc/grid-security/certificates");
  res = v.Import(binary);

  if (res)
    cout << "------------------" << endl << binary << "-------" << endl << flush;
  else
    cerr << "Error!" << endl << flush;
  if (!res) {
    cerr << "err: " << retmsg[v.error] << endl;
  }


  cerr << "TEST 2" << endl;
  /*
   * Get Data from real certificate.
   */

  d.data.clear();

  BIO *in = NULL;
  X509 *x = NULL;

  in = BIO_new(BIO_s_file());
  if (in) {
    if (BIO_read_filename(in, "/tmp/x509up_u500") > 0) {
      x = PEM_read_bio_X509(in, NULL, 0, NULL);
      res = d.Retrieve(x, NULL, RECURSE_NONE);

      if (res)
	print(d);
      else
	cerr << "ERROR!" << endl << flush;
    }
  }

  if (!res) {
    cerr << "err: " << retmsg[d.error] << endl;
  }
  cerr << flush;
  
  exit(0);
}
