#include "../../std.h"

#define GRIDFTP_PLUGIN
#include "gaclplugin.h"
#include "../../config/conf.h"
#include "../userspec.h"
#include "../names.h"
#include "../misc.h"
#include "../../misc/inttostring.h"
#include "../../misc/canonical_dir.h"
#include "../../misc/log_time.h"

#ifndef HAVE_STRERROR_R
char * strerror_r (int errnum, char * buf, size_t buflen) {
  char * estring = strerror (errnum);
  strncpy (buf, estring, buflen);
  return buf;
}
#endif

#define  ADD_SUBSTITUTE_VALUE(nam,val) \
  if(val) { \
    struct _GACLnamevalue *s = \
       (struct _GACLnamevalue *)malloc(sizeof(struct _GACLnamevalue)); \
    if(s) { \
      s->next=subst; subst=s; \
      subst->name=strdup(nam); \
      subst->value=strdup(val); \
    }; \
  }


GACLPlugin::GACLPlugin(istream &cfile,userspec_t &u) {
  data_file=-1;
  subst=NULL;
  /* read configuration */
  char xmlbuf[ACL_MAX_SIZE+1];
  int  xmllen = 0;
  int skip=0;
  acl=NULL;
  subject=u.user.DN();
  user=&(u.user);
  subst=(struct _GACLnamevalue *)malloc(sizeof(struct _GACLnamevalue));
  if(subst) {
    subst->next=NULL;
    subst->name=strdup("subject");
    subst->value=strdup(subject.c_str());
  };
  ADD_SUBSTITUTE_VALUE("vo",(u.user.default_vo()));
  ADD_SUBSTITUTE_VALUE("role",u.user.default_role());
  ADD_SUBSTITUTE_VALUE("capability",u.user.default_capability());
  ADD_SUBSTITUTE_VALUE("group",u.user.default_vgroup());
  ADD_SUBSTITUTE_VALUE("voms",u.user.default_voms());
  if(!cfile.eof()) {
    istream_readline(cfile,xmlbuf,sizeof(xmlbuf));
    basepath=xmlbuf;
  };
  for(;!cfile.eof();) {
    char* p = xmlbuf+xmllen;
    char* command;
    if(skip) { p=xmlbuf; xmllen=0; };
    istream_readline(cfile,p,sizeof(xmlbuf)-xmllen);
    for(;*p;p++) if(!isspace(*p)) break;
    command=p;
    for(;*p;p++) if(isspace(*p)) break;
    if(((p-command)==3) && (strncmp(command,"end",3)==0)) {
      xmlbuf[xmllen]=0;
      break;
    };
    xmllen=strlen(xmlbuf);
    if((sizeof(xmlbuf)-xmllen) <= 1) skip=1;
  };
  if(xmllen) {
    acl=GACLacquireAcl(xmlbuf);
  } else {
    acl=NULL;
  };
}

GACLPlugin::~GACLPlugin(void) {
  for(;subst;) {
    struct _GACLnamevalue *s = subst->next;
    if(subst->name) free(subst->name);
    if(subst->value) free(subst->value);
    free(subst);
    subst=s;
  };
  if(acl) GACLfreeAcl(acl);
}

/* name must be absolute path */
/* make directories out of scope of mount dir */
static int makedirs(string &name) {
  struct stat64 st;
  if(stat64(name.c_str(),&st) == 0) {
    if(S_ISDIR(st.st_mode)) return 0;
    return 1;
  };
  int n=1;
  for(;;) {
    if(n >= name.length()) break;
    n=name.find('/',n); if(n==-1) n=name.length();
    string dname=name.substr(0,n); n++;
    if(stat64(dname.c_str(),&st) == 0) {
      if(S_ISDIR(st.st_mode)) continue;
      return 1;
    };
    if(mkdir(dname.c_str(),S_IRWXU | S_IRWXG | S_IRWXO) == 0) continue;
    char* errmsg;
    char errmgsbuf[256];
    errmsg=strerror_r(errno,errmgsbuf,sizeof(errmgsbuf));
    olog<<"mkdir failed: "<<errmsg<<endl;
    return 1;
  };
  return 0;
}

int GACLPlugin::removedir(string &name) {
  const char* basename = get_last_name(name.c_str());
  if(strncmp(basename,".gacl-",6) == 0) return 1;
  string dirname = basepath+"/"+name;
  GACLperm perm=GACLtestFileAclForVOMS(dirname.c_str(),subject.c_str(),user->voms());
  if(!GACLhasWrite(perm)) return 1;
  struct stat64 st;
  if(stat64(dirname.c_str(),&st) != 0) return 1;
  if(!S_ISDIR(st.st_mode)) return 1;
  if(::remove(dirname.c_str()) != 0) return 1;
  GACLdeleteFileAcl(dirname.c_str());
  return 0;
}

int GACLPlugin::removefile(string &name) {
  const char* basename = get_last_name(name.c_str());
  if(strncmp(basename,".gacl-",6) == 0) return 1;
  string filename = basepath+"/"+name;
  GACLperm perm=GACLtestFileAclForVOMS(filename.c_str(),subject.c_str(),user->voms());
  if(!GACLhasWrite(perm)) return 1;
  struct stat64 st;
  if(stat64(filename.c_str(),&st) != 0) return 1;
  if(!S_ISREG(st.st_mode)) return 1;
  if(::remove(filename.c_str()) != 0) return 1;
  GACLdeleteFileAcl(filename.c_str());
  return 0;
}

int GACLPlugin::makedir(string &name) {
  string dname=basepath;
  if(makedirs(dname) != 0) { /* can't make mount point */
    olog << "Warning: mount point " << dname << " creation failed." << endl;
    return 1; 
  };
  /* now go through rest of directories */
  int n = 0;
  for(;;) {
    /* default acl */
    string gname=dname+"/.gacl";
    if(!GACLsaveSubstituted(acl,subst,gname.c_str())) {
      struct stat st;
      if(stat(gname.c_str(),&st) != 0) return 1;
      if(!S_ISREG(st.st_mode)) return 1;
    };
    if(n >= name.length()) break;
    n=name.find('/',n); if(n==-1) n=name.length();
    dname=basepath+"/"+name.substr(0,n);
    gname=basepath+"/.gacl-"+name.substr(0,n);
    n++;
    if(!GACLsaveSubstituted(acl,subst,gname.c_str())) {
      struct stat st;
      if(stat(gname.c_str(),&st) != 0) return 1;
      if(!S_ISREG(st.st_mode)) return 1;
    };
  };
  return 0;
}

int GACLPlugin::open(const char* name,open_modes mode,unsigned long long int size) {
  olog<<"plugin(gacl): open: "<<name<<endl;
  file_mode=file_access_none;
  string filename = basepath+"/"+name;
  string dname=name; if(!remove_last_name(dname)) { return 1; };
  const char* basename = get_last_name(name);
  bool acl_request = (strncmp(basename,".gacl-",6) == 0);
  if(acl_request) filename=basepath+"/"+(name+6);
  if( mode == GRIDFTP_OPEN_RETRIEVE ) {  /* open for reading */
    GACLperm perm=GACLtestFileAclForVOMS(filename.c_str(),subject.c_str(),user->voms());
    if(acl_request) {
      if(!GACLhasAdmin(perm)) return 1;
      string gname = basepath+"/"+name;
      int h=::open(gname.c_str(),O_RDONLY);
      if(h==-1) return 1; /* TODO - open default file */
      int l = 0;
      for(;;) {
        int ll = sizeof(acl_buf)-l;
        if(ll<=0) break;
        ll = ::read(h,acl_buf+l,ll);
        if(ll == -1) { close(h); return 1; };
        if(ll == 0) break;
        l+=ll; 
      };
      close(h);
      if(l>=sizeof(acl_buf)) return 1;
      acl_buf[l]=0; acl_length=l;
      file_mode=file_access_read_acl;
      file_name=filename;
      return 0;   
    };
    if(GACLhasRead(perm)) {
      data_file=::open64(filename.c_str(),O_RDONLY);
      if(data_file == -1) return 1;
      file_mode=file_access_read;
      file_name=filename;
      return 0;
    };
    return 1;
  }
  else if( mode == GRIDFTP_OPEN_STORE ) { /* open for writing */
    if(acl_request) {
      GACLperm perm=GACLtestFileAclForVOMS(filename.c_str(),subject.c_str(),user->voms());
      if(!GACLhasAdmin(perm)) return 1;
      struct stat64 st; /* make sure file exists */
      if(stat64(filename.c_str(),&st) != 0) return 1;
      if(!S_ISREG(st.st_mode)) return 1;
      memset(acl_buf,0,sizeof(acl_buf));
      file_mode=file_access_write_acl;
      file_name=filename;
      return 0;
    };
    /* first check if file exists */
    struct stat64 st;
    if(stat64(filename.c_str(),&st) == 0) {  /* overwrite */
      if(S_ISREG(st.st_mode)) {
        GACLperm perm=GACLtestFileAclForVOMS(filename.c_str(),subject.c_str(),user->voms());
        if(GACLhasWrite(perm)) {
          if(size > 0) {
            struct statfs dst;
            if(statfs(filename.c_str(),&dst) == 0) {
              if(size > ((dst.f_bfree*dst.f_bsize) + st.st_size)) {
                olog<<"Not enough space to store file"<<endl;
                return 1;
              };
            };
          };
          data_file=::open64(filename.c_str(),O_WRONLY);
          if(data_file == -1) return 1;
          file_mode=file_access_overwrite;
          file_name=filename;
          truncate(file_name.c_str(),0);
          return 0;
        };
        return 1;
      }
      else if(S_ISDIR(st.st_mode)) { /* it's a directory */
        return 1;
      };
      return 1;
    }
    else { /* no such object in filesystem */
      GACLperm perm=GACLtestFileAclForVOMS(filename.c_str(),subject.c_str(),user->voms());
      if(GACLhasWrite(perm)) {
        if(makedir(dname) != 0) return 1;
        if(size > 0) {
          struct statfs dst;
          if(statfs(filename.c_str(),&dst) == 0) {
            if(size > (dst.f_bfree*dst.f_bsize)) {
              olog<<"Not enough space to store file"<<endl;
              return 1;
            };
          };
        };
        data_file=::open64(filename.c_str(),O_WRONLY | O_CREAT | O_EXCL,S_IRUSR | S_IWUSR);
        if(data_file == -1) return 1;
        file_name=filename;
        file_mode=file_access_create;
        return 0;
      };
      return 1;
    };
  }
  else {
    olog << "Warning: unknown open mode " << mode << endl;
    return 1;
  };
  return 1;
}

int GACLPlugin::close(bool eof) {
  if((file_mode==file_access_read_acl) || (file_mode==file_access_write_acl)) { 
    if(eof) {
      if(file_mode==file_access_write_acl) {
        file_mode=file_access_none;
        GACLacl* acl = GACLacquireAcl(acl_buf);
        if(acl == NULL) return 1;
        int n=file_name.rfind('/'); if(n==string::npos) n=0;
        string gname=file_name;
        gname.insert(n+1,".gacl-");
        if(!GACLsaveAcl((char*)(gname.c_str()),acl)) {
          GACLfreeAcl(acl);
          return 1;
        };
        GACLfreeAcl(acl);
        return 0;
      };
      file_mode=file_access_none;
      return 0;
    }; 
    file_mode=file_access_none;
    return 0;
  };
  if(data_file != -1) {
    if(eof) {
      ::close(data_file);
      if((file_mode==file_access_create) || 
         (file_mode==file_access_overwrite)) { 
        int n=file_name.rfind('/'); if(n==string::npos) n=0;
        if(acl) { // store per-file default acl only if specified
          string gname=file_name;
          gname.insert(n+1,".gacl-");
          GACLsaveSubstituted(acl,subst,gname.c_str());
        };
      };
    }
    else {  /* file was not transfered properly */
      if((file_mode==file_access_create) || 
         (file_mode==file_access_overwrite)) { /* destroy file */
        ::close(data_file);
        ::unlink(file_name.c_str());
      };
    };
  };
  file_mode=file_access_none;
  return 0;
}

int GACLPlugin::read(unsigned char *buf,unsigned long long int offset,unsigned long long int *size) {
  if(file_mode==file_access_read_acl) {
    if(offset >= acl_length) { (*size)=0; return 0; };
    int l = (acl_length - offset);
    (*size)=l;
    memcpy(buf,acl_buf+offset,l);
    return 0;
  };
  ssize_t l;
  if(data_file == -1) return 1;
  if(lseek64(data_file,offset,SEEK_SET) != offset) {
    (*size)=0; return 0; /* can't read anymore */
  };
  if((l=::read(data_file,buf,(*size))) == -1) {
    olog <<"Warning: error while reading file"<<endl;
    (*size)=0; return 1;
  };
  (*size)=l;
  return 0;
}

int GACLPlugin::write(unsigned char *buf,unsigned long long int offset,unsigned long long int size) {
  if(file_mode==file_access_write_acl) {
    if(offset >= (sizeof(acl_buf)-1)) { return 1; };
    if((offset+size) > (sizeof(acl_buf)-1)) { return 1; };
    memcpy(acl_buf+offset,buf,size);
    return 0;
  };
  ssize_t l;
  size_t ll;
//  olog << "plugin: write\n";
  if(data_file == -1) return 1;
  if(lseek64(data_file,offset,SEEK_SET) != offset) {
    perror("lseek");
    return 1; /* can't write at that position */
  };
  for(ll=0;ll<size;ll+=l) {
    if((l=::write(data_file,buf+ll,size-ll)) == -1) {
      perror("write");
      return 1;
    };
    if(l==0) olog << "Warning: zero bytes written to file" << endl;
  };
  return 0;
}

int GACLPlugin::readdir(const char* name,list<DirEntry> &dir_list,DirEntry::object_info_level mode) {
  string dirname = basepath+"/"+name;
  /* check if allowed to read this directory */
  GACLperm perm=GACLtestFileAclForVOMS(dirname.c_str(),subject.c_str(),user->voms());
  if(!GACLhasList(perm)) return 1;
  struct stat st;
  if(stat(dirname.c_str(),&st) != 0) return 1;
  if(S_ISDIR(st.st_mode)) {
    DIR* d=::opendir(dirname.c_str());
    if(d == NULL) { return 1; }; /* maybe return ? */
    struct dirent *de;
    for(;;) {
      de=::readdir(d);
      if(de == NULL) break;
      if((!strcmp(de->d_name,".")) || 
         (!strcmp(de->d_name,"..")) ||
         (!strcmp(de->d_name,".gacl")) ||
         (!strncmp(de->d_name,".gacl-",6))
      ) continue;
      DirEntry dent(true,de->d_name); // treat it as file by default
      bool is_manageable = fill_object_info(dent,dirname,mode);
      if(is_manageable) {
        dir_list.push_back(dent);
      };
    };
    ::closedir(d);
    return 0;
  }
  else if(S_ISREG(st.st_mode)) {
    DirEntry dent(true,"");
    bool is_manageable = fill_object_info(dent,name,mode);
    if(is_manageable) {
      dir_list.push_back(dent);
      return -1;
    };
  };
  return 1;
}

/* checkdir is allowed to change dirname to show actual target of cd */
int GACLPlugin::checkdir(string &name) {
  string dirname = basepath+"/"+name;
  GACLperm perm=GACLtestFileAclForVOMS(dirname.c_str(),subject.c_str(),user->voms());
  if(!GACLhasList(perm)) return 1;
  struct stat st;
  if(stat(dirname.c_str(),&st) != 0) return 1;
  if(!S_ISDIR(st.st_mode)) return 1;
  return 0;
}

int GACLPlugin::checkfile(string &name,DirEntry &info,DirEntry::object_info_level mode) {
  const char* basename = get_last_name(name.c_str());
  if(strncmp(basename,".gacl-",6) == 0) { /* hack - any acl already exists */
    DirEntry dent(true,basename);
    info=dent;
    return 0;
  };
  string filename = basepath+"/"+name;
  GACLperm perm=GACLtestFileAclForVOMS(filename.c_str(),subject.c_str(),user->voms());
  if(!GACLhasList(perm)) return 1;
  DirEntry dent(true,get_last_name(filename.c_str()));
  string dirname = filename; remove_last_name(dirname);
  bool is_manageable = fill_object_info(dent,dirname,mode);
  if(!is_manageable) {
    return 1;
  };
  info=dent;
  return 0;
}

bool GACLPlugin::fill_object_info(DirEntry &dent,string dirname,
                      DirEntry::object_info_level mode) {
  bool is_manageable = true;
  if(mode != DirEntry::minimal_object_info) {
    string ffname = dirname;
    if(dent.name.length() != 0) ffname+="/"+dent.name;
    struct stat64 st;
    if(stat64(ffname.c_str(),&st) != 0) { is_manageable=false; }
    else {
      // TODO: treat special files (not regular) properly. (how?)
      if((!S_ISREG(st.st_mode)) && (!S_ISDIR(st.st_mode))) is_manageable=false;
    };
    if(is_manageable) {
      dent.uid=st.st_uid;
      dent.gid=st.st_gid;
      dent.changed=st.st_ctime;
      dent.modified=st.st_mtime;
      dent.is_file=S_ISREG(st.st_mode);
      dent.size=st.st_size;
      if(mode != DirEntry::basic_object_info) {
        GACLperm perm=GACLtestFileAclForVOMS(ffname.c_str(),subject.c_str(),user->voms());
        if(dent.is_file) {
          if(GACLhasWrite(perm)) {
            dent.may_delete=true; dent.may_write=true; dent.may_append=true;
          };
          if(GACLhasRead(perm)) {
            dent.may_read=true;
          };
        } else {
          if(GACLhasWrite(perm)) {
            dent.may_delete=true;
            dent.may_create=true;
            dent.may_mkdir=true;
            dent.may_purge=true;
          };
          if(GACLhasList(perm)) {
            dent.may_chdir=true;
            dent.may_dirlist=true;
          };
        };
      };
    };
  };
  return is_manageable;
}

